#include <stdio.h>
#include <stdlib.h>
#include <omp.h>
#include "mpi.h"

/*  This code is an implementation of the slope one method  */

#define MASTER 0
#define FROM_MASTER 1
#define FROM_WORKER 2
#define DIETAG 3
#define TRUE 1

#define ITEMS 3952
#define users 1500
#define users2 100 /* How many rows/users of ratings array we send each time to workers  */
#define records 248579

float predictions(int *ratings,int *denominator,float *deviation,int user,int item);
float weightedpredictions(int *ratings,int *denominator,float *deviation,int user,int item);

int main ( argc , argv )
int argc;
char **argv;
{
int size,rank;
int counter=0; /* counts how many times we have send data to workers */
int active=0; /* How many processes are active */
int tag;
int sender=0;
MPI_Status status;
MPI_Init ( &argc , &argv );
MPI_Comm_rank( MPI_COMM_WORLD, &rank);
MPI_Comm_size( MPI_COMM_WORLD, &size);
int buffer[records][3];
int i=0;
int j=0;
int k=0;
int q=0;
int testuser;
int user;
int item;
int sth=1;

/* ###### Variables for time calculation ###### */
double total_time_start,total_time_end,comm_time_start,comm_time_end,IO_time_start,IO_time_end,calc_time_start,calc_time_end;
double comm_time=0.0;
double comm_time_worker=0.0;
double IO_time=0.0;
double calc_time=0.0;
double calc_time_worker=0.0;
double pred_time_start,pred_time_end,wpred_time_start,wpred_time_end,time_IO_start,time_IO_end;
double pred_time=0.0;
double wpred_time=0.0;
double pre_proc=0.0;
double time_IO=0.0;
double time_IOw=0.0;

total_time_start=MPI_Wtime();
/* ###### All processes including master - Memory Allocation ###### */

int *ratings,*numerator,*denominator,*testratings,*denominator2;
float *deviation,*deviation2;

ratings=(int *)calloc( users*ITEMS ,sizeof(int) );
if (ratings==NULL) {
printf("Memory allocation for ratings failed \n");
}

testratings=(int *)calloc( users2*ITEMS ,sizeof(int) );
if (testratings==NULL) {
printf("Memory allocation for testratings failed \n");
}

numerator=(int*)calloc(ITEMS*ITEMS ,sizeof(int) );
if(numerator==NULL) {
printf("Memory allocation for numerator failed \n");
}

denominator=(int*)calloc( ITEMS*ITEMS ,sizeof(int) );
if(denominator==NULL) {
printf("Memory allocation for denominator failed \n");
}

denominator2=(int*)calloc( ITEMS*ITEMS ,sizeof(int) );
if(denominator2==NULL) {
printf("Memory allocation for denominator2 failed \n");
}

deviation=(float*)calloc( ITEMS*ITEMS ,sizeof(float) );
if(deviation==NULL) {	
printf("Memory allocation for deviation failed \n");
}

deviation2=(float*)calloc( ITEMS*ITEMS ,sizeof(float) );
if(deviation2==NULL) {
printf("Memory allocation for deviation2 failed \n");
}


/* ###### ------ ###### */

/* ######  Master Code ###### */
if (rank==0) 
{

/* ###### Opening file ###### */
IO_time_start=MPI_Wtime();
FILE *fpe;
fpe=fopen("data1","r");
if(fpe==NULL)
{
printf("File data1 cannot be opened. \n");
MPI_Finalize();
return -1;
}

/* ###### Reading data from text file and set buffer and ratings ###### */
/* ###### Sending parts of ratings to workers ###### */
int user,item,rating,ts;

	for(j=0;j<records;j++)
	{
		fscanf(fpe,"%i\t%i\t%i\t%i ",&user,&item,&rating,&ts);
		buffer[j][0]=user;
		buffer[j][1]=item;
		buffer[j][2]=rating;
	}
fclose(fpe);
IO_time_end=MPI_Wtime();
IO_time = IO_time + IO_time_end - IO_time_start;
/* ###### Set Ratings ###### */
calc_time_start=MPI_Wtime();
#pragma omp parallel private(i,j,k,testuser)
{
#pragma omp for 
for(i=0;i<records;i++) {
	testuser=buffer[i][0];
	j=buffer[i][1];
	k=(testuser*ITEMS)-ITEMS+j-1;
	ratings[k]=buffer[i][2];
	}
} /* end of parallel section */
calc_time_end=MPI_Wtime();
calc_time = calc_time + calc_time_end - calc_time_start;
 /* ######  
//print ratings
printf("### Ratings ### \n");
k=0;
for(i=0;i<users;i++){
	for(j=0;j<ITEMS;j++) {
	printf(" ratings[%i][%i]=%i \n",i+1,j+1,ratings[k]);
	k++;
	}
}
 ###### */

/* ###### Sending testratings to workers ###### */
for(i=1;i<size;i++)
{
	calc_time_start=MPI_Wtime();
	q=counter*ITEMS*users2;

	#pragma omp parallel private(j)
	{
	#pragma omp for 
	for(j=0;j<users2*ITEMS ;j++)
	{
		testratings[j]=ratings[q+j];
		
	}
	} /* end of parallel section */
	calc_time_end=MPI_Wtime();
	calc_time = calc_time + calc_time_end - calc_time_start;
	//MPI_Send(&counter,1,MPI_INT,i,6,MPI_COMM_WORLD);
	comm_time_start=MPI_Wtime();
	MPI_Send(testratings, users2*ITEMS, MPI_INT, i, FROM_MASTER,MPI_COMM_WORLD);
	comm_time_end=MPI_Wtime();
	comm_time = comm_time  +  comm_time_end  -  comm_time_start ;
	counter++;
	active++;
}	

/* ######  //Print buffer 
for(i=0;i<records;i++) {
	for(j=0;j<3;j++) {
	printf("buffer[%i][%i]=%i  ",i,j,buffer[i][j]);
	 }
	printf ("\n");
 }
###### */

do
{
/* ###### Master receiving sth from workers ###### */
comm_time_start=MPI_Wtime();
MPI_Recv(&sth,1,MPI_INT,MPI_ANY_SOURCE,7,MPI_COMM_WORLD,&status);
comm_time_end=MPI_Wtime();
comm_time = comm_time  +  comm_time_end  -  comm_time_start ;
active--;
sender=(status.MPI_SOURCE);
/* ###### Master sending next data or dietag to workers ###### */
if(counter*users2<=users)
{
	calc_time_start=MPI_Wtime();
	#pragma omp parallel private(j)
		{
		#pragma omp for 
		for(j=0;j<users2*ITEMS ;j++)
		{
		testratings[j]=ratings[counter*ITEMS*users2+j];
		}
	} /* end of parallel section */
	calc_time_end=MPI_Wtime();
	calc_time = calc_time + calc_time_end - calc_time_start;		
	//MPI_Send(&counter,1,MPI_INT,sender,6,MPI_COMM_WORLD);
	comm_time_start=MPI_Wtime();
	MPI_Send(testratings, users2*ITEMS, MPI_INT, sender, FROM_MASTER,MPI_COMM_WORLD);
	comm_time_end=MPI_Wtime();
	comm_time = comm_time  +  comm_time_end  -  comm_time_start ;
	counter++;
	active++;
}
else if ( counter*users2>users )
{
	tag=DIETAG;
	comm_time_start=MPI_Wtime();
	MPI_Send(testratings,users2*ITEMS, MPI_INT, sender, tag,MPI_COMM_WORLD);
	comm_time_end=MPI_Wtime();
	comm_time = comm_time  +  comm_time_end  -  comm_time_start ;
	counter++;
	/* ###### Master receiving results from workers ###### */
	comm_time_start=MPI_Wtime();
	MPI_Recv(denominator2,ITEMS*ITEMS,MPI_INT,MPI_ANY_SOURCE,4,MPI_COMM_WORLD,&status);
	MPI_Recv(deviation2,ITEMS*ITEMS,MPI_FLOAT,MPI_ANY_SOURCE,5,MPI_COMM_WORLD,&status );
	comm_time_end=MPI_Wtime();
	comm_time = comm_time  +  comm_time_end  -  comm_time_start ;
	/* ###### Master placing results to right position ###### */
	calc_time_start=MPI_Wtime();
	#pragma omp parallel private(i,j)
	{
	#pragma omp for 
	for(i=0;i<ITEMS;i++)
	{
		for(j=0;j<ITEMS;j++)
		{
		denominator[ i*ITEMS+j]=denominator[ i*ITEMS+j ]+denominator2[ i*ITEMS+j ];
		deviation[ i*ITEMS+j ]=deviation[ i*ITEMS+j]+deviation2[ i*ITEMS+j ];
		}
	}
	} /* end of parallel section */
	calc_time_end=MPI_Wtime();
	calc_time = calc_time + calc_time_end - calc_time_start;	
}

}while(active>0);
calc_time_start=MPI_Wtime();
#pragma omp parallel private(i,j,k)
{
/* ###### Calculate Deviation ###### */

//print deviation
//printf("### Deviation ### \n");
k=0;
#pragma omp for 
	for(i=1;i<=ITEMS;i++)
	{
	k=(i-1)*ITEMS;
		for(j=1;j<=ITEMS;j++)
		{
			//calculate deviation
			if(denominator[k]==0)
			{
			deviation[k]=0;	
			}
			else
			{
				deviation[k]=deviation[k]/denominator[k];
			}
//			printf(" deviation[%i][%i]=%f \t",i,j,deviation[k]);
			k++;
		}
//	printf("\n");
	}

/* ###### ------ ###### */

} /* end of parallel section */
calc_time_end=MPI_Wtime();
calc_time = calc_time + calc_time_end - calc_time_start;
pre_proc=calc_time_end - total_time_start;
/* ###### Predictions ###### */
//printf("############\n");

float pred;
//pred=predictions(ratings,denominator,deviation,1,5);
//printf("The prediction for user 1 and item 5 is:%f \n",pred);
//pred=predictions(ratings,denominator,deviation,5,427);
//printf("The prediction for user 5 and item 427 is:%f \n",pred);
//pred=weightedpredictions(ratings,denominator,deviation,5,3);
//printf("The weighted prediction for user 5 and item 3 is:%f \n",pred);
//pred=weightedpredictions(ratings,denominator,deviation,3,357);
//printf("The weighted prediction for user 3 and item 357 is:%f \n",pred);
/* ###### Create file predictions ###### */
IO_time_start=MPI_Wtime();
FILE *fp;
fp=fopen("predictions.txt","w");
if(fp==NULL)
{
	printf("File predictions could not be opened. \n");
	MPI_Finalize();
	return -1;
}

fprintf(fp,"User\tItem\t Rating\n");
IO_time_end=MPI_Wtime();
IO_time=IO_time+IO_time_end-IO_time_start;
pred_time_start=MPI_Wtime();
#pragma omp parallel private(i,user,item)
{
#pragma omp for
for(i=0;i<users*ITEMS;i++)
{
	if(ratings[i]==0)
	{
		user=i/ITEMS+1;
		item=i%ITEMS+1;
		pred=predictions(ratings,denominator,deviation,user,item);
		time_IO_start=MPI_Wtime();
		fprintf(fp,"%d\t%d\t%f\n",user,item,pred);
		time_IO_end=MPI_Wtime();
		time_IO=time_IO+time_IO_end-time_IO_start;
		IO_time=IO_time+time_IO_end-time_IO_start;
	}
}
} /* End of parallel section */
pred_time_end=MPI_Wtime();
pred_time=pred_time_end-pred_time_start-time_IO;
IO_time_start=MPI_Wtime();
fclose(fp);
/* ###### Create file weightedpredictions ###### */
FILE *fpw;
fpw=fopen("weightedpredictions.txt","w");
if(fpw==NULL)
{
	printf("File weightedpredictions could not be opened. \n");
	MPI_Finalize();
	return -1;
}

fprintf(fpw,"User\tItem\t Rating\n");
IO_time_end=MPI_Wtime();
IO_time=IO_time+IO_time_end-IO_time_start;
wpred_time_start=MPI_Wtime();
#pragma omp parallel private(i,user,item)
{
#pragma omp for
for(i=0;i<users*ITEMS;i++)
{
	if(ratings[i]==0)
	{
		user=i/ITEMS+1;
		item=i%ITEMS+1;
		pred=weightedpredictions(ratings,denominator,deviation,user,item);
		time_IO_start=MPI_Wtime();
		fprintf(fpw,"%d\t%d\t%f\n",user,item,pred);
		time_IO_end=MPI_Wtime();
		time_IOw=time_IOw+time_IO_end-time_IO_start;
		IO_time=IO_time+time_IO_end-time_IO_start;
	}
}
} /* End of parallel section */
wpred_time_end=MPI_Wtime();
wpred_time=wpred_time_end-wpred_time_start-time_IOw;
IO_time_start=MPI_Wtime();
fclose(fpw);
IO_time_end=MPI_Wtime();
IO_time = IO_time + IO_time_end - IO_time_start ;

/* ###### ------ ###### */
/* ###### ------ ###### */
} /*  End of Master code */

else
{
/* ###### Workers code ###### */


int w=0;
int m=0;

while(TRUE)
{


//MPI_Recv(&counter,1,MPI_INT,MASTER,6,MPI_COMM_WORLD,&status);
comm_time_start=MPI_Wtime();
MPI_Recv(testratings, users2*ITEMS, MPI_INT,MASTER,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
comm_time_end=MPI_Wtime();
comm_time_worker = comm_time_worker  +  comm_time_end  -  comm_time_start ;


/* ####### Check the tag of the received message ###### */
tag=status.MPI_TAG;
if(tag==DIETAG){
comm_time_start=MPI_Wtime();
MPI_Send(denominator2,ITEMS*ITEMS,MPI_INT,MASTER,4,MPI_COMM_WORLD);
MPI_Send(deviation2,ITEMS*ITEMS,MPI_FLOAT,MASTER,5,MPI_COMM_WORLD);
comm_time_end=MPI_Wtime();
comm_time_worker = comm_time_worker  +  comm_time_end  -  comm_time_start ;
break;
}
calc_time_start=MPI_Wtime();
/* ###### Set numerator(differences matrix) - denominator(frequencies matrix) ###### */
/* ###### Set (pre)deviation ###### */

int x=0;
int z=0;
#pragma omp parallel private(i,j,k,testuser,x,w,m,z) firstprivate(numerator)
{

#pragma omp for 
for(testuser=1;testuser<=users2;testuser++) 
{ 
	x=(testuser-1)*ITEMS;
	//k=(testuser-1)*ITEMS*ITEMS;
	k=0;
	w=0;
	m=0;
	for(i=1;i<=ITEMS;i++) {
		for(j=1;j<=ITEMS;j++) {
		if(i==j){
			if( (testratings[x+i-1]==0) || (testratings[x+j-1]==0) ) {
			numerator[k]=10;
			}
			else {	
			numerator[k]=0;
			}
			k++;
			}
		else {  
			if( (testratings[x+i-1]==0) || (testratings[x+j-1]==0) ) {
			numerator[k]=10;
			}
			else {
			numerator[k]=testratings[x+i-1]-testratings[x+j-1];
			}
			k++;
			}
//			printf(" numerator[%i][%i][%i]=%i \t",testuser,i,j,numerator[k-1]);
			//calculate deviation's numerator	
			if(numerator[k-1]!=10){
			deviation2[m]=deviation2[m]+numerator[k-1];
			}
			m++;

	
		} //end for j
//	printf("\n");	
	if( testratings[x+i-1]!=0 ) {
				w=w+i-1;
				denominator2[w]=denominator2[w]+1;
				w++;
				for(j=i+1;j<=ITEMS;j++) {
					if( testratings[x+j-1]!=0 ) {
				 		denominator2[w]=denominator2[w]+1;
					}
				w++;
				}
			}
			else if (testratings[x+i-1]==0){
				w=w+i-1;
				w++;
				for(j=i+1;j<=ITEMS;j++) {
				w++;
				}
			}	
	}
	for(z=0;z<ITEMS*ITEMS;z++)
	{
		numerator[z]=0;
	}

}



// calculate symmetric values of denominator

#pragma omp for 
for(i=1;i<ITEMS;i++){
	z=i*ITEMS;
	denominator2[z]=denominator2[i];
	for(k=0;k<=i-1;k++){
		z++;
		denominator2[z]=denominator2[z-(i-1)*(ITEMS-1)+(k*(ITEMS-1))];
	}
}

} /* end of parallel section */
calc_time_end=MPI_Wtime();
calc_time_worker = calc_time + calc_time_end - calc_time_start;
 /* ######
//print denominator
printf("### Denominator (Frequencies) ### \n");
k=0;
	for(i=1;i<=ITEMS;i++) {
		for(j=1;j<=ITEMS;j++) {
				printf(" denominator[%i][%i]=%i \t",i,j,denominator[k]);
				k++;
				}
			printf("\n");
			}
 ###### */

/* ###### ------ ###### */
comm_time_start=MPI_Wtime();
MPI_Send(&sth,1,MPI_INT,MASTER,7,MPI_COMM_WORLD);
comm_time_end=MPI_Wtime();
comm_time_worker = comm_time_worker  +  comm_time_end  -  comm_time_start ;


}  /* end of while */
printf("############\n");
printf("Worker with rank: %d Communication time: %f\n",rank,comm_time_worker);
printf("Worker with rank: %d Computation time: %f \n",rank,calc_time_worker);

}/*  End of Workers code */

/* ###### ------ ###### */



free(ratings);

free(denominator);
free(numerator);
free(deviation);
free(testratings);
free(denominator2);
free(deviation2);
total_time_end=MPI_Wtime();
/* ###### Printing time results ###### */
if(rank==0)
{
printf("############\n");
printf("Total execution time is: %f \n",total_time_end-total_time_start);
printf("Communication time on master's code is: %f \n",comm_time);
printf("Total IO time is: %f \n",IO_time);
printf("Computation time on master's code is: %f \n",calc_time);
printf("Data preprocess time is %f \n",pre_proc);
printf("Predictions time is: %f \n",pred_time);
printf("Weighted predictions time is: %f \n",wpred_time);
printf("########################\n");
}
MPI_Finalize ();
return 0;

}
 
 
 /* ###### Function Predictions ###### */

float predictions(int *ratings,int *denominator,float *deviation,int user,int item){

float prediction=0.0;
float u=0.0;
int count=0;
float sum=0.0;
int i=0;
int cardRi=0;
float Sdev=0.0;
int item2;

#pragma omp parallel for default(none)  shared(user,ratings) \
		reduction(+:sum,count)  
for(i=((user*ITEMS)-ITEMS);i<user*ITEMS;i++){
	if(ratings[i]!=0){
		count++;
		sum=sum+ratings[i];
		}
	}
	
/* end of parallel  for section and reduction */
u=sum/count;	
if(ratings[(user*ITEMS)+item-(ITEMS+1)]==0){
	cardRi=count;
	}
	else{
	cardRi=count-1;
	}
#pragma omp parallel for default(none)  shared(user,deviation,ratings,denominator,item) \
		reduction(+:Sdev)  	
for(item2=1;item2<=ITEMS;item2++){
	if(item2!=item && ratings[user*ITEMS+item2-(ITEMS+1)]!=0 && denominator[(item-1)*ITEMS+item2-1]!=0){
			Sdev=Sdev+deviation[(item-1)*ITEMS+item2-1];
		}
	}
/* end of parallel  for section and reduction */	
if(cardRi==0 || Sdev==0){
//printf("Prediction not available for user %i and item %i \n",user,item);
}
else{
prediction=u+Sdev/cardRi;
}
return prediction;
}

/* ###### ------ ###### */

/* ###### Function Weighted Predictions ###### */

float weightedpredictions(int *ratings,int *denominator,float *deviation,int user,int item){

float prediction=0.0;
float S1=0.0;
int S2=0;
int item2;
#pragma omp parallel for default(none)  shared(user,deviation,ratings,denominator,item) \
		reduction(+:S1,S2) 
for(item2=1;item2<=ITEMS;item2++){
	if(item2!=item && ratings[user*ITEMS+item2-(ITEMS+1)]!=0 ){
			S1=S1+(deviation[(item-1)*ITEMS+item2-1]+ratings[user*ITEMS+item2-(ITEMS+1)])*denominator[(item-1)*ITEMS+item2-1];
			S2=S2+denominator[(item-1)*ITEMS+item2-1];
		}
	}
/* end of parallel  for section and reduction */	
if(S2==0){
//printf("Prediction not available for user %i and item %i \n",user,item);
}
else{
prediction=S1/S2;
}
return prediction;
}

/* ###### ------ ###### */
